{
  "type": "record",
  "namespace": "yelp.data_pipeline",
  "name": "monitoring_message",
  "doc": "Monitoring message that counts the number of messages that are produced/consumed by specific client in a given time frame",
  "fields": [
    {
      "name": "topic",
      "type": "string",
      "doc": "topic associated with the messages produced/consumed by client"
    },
    {
      "name": "client_name",
      "type": "string",
      "doc": "name of the client that is publishing/consuming messages"
    },
    {
      "name": "client_type",
      "type": {
        "name": "client_type",
        "doc": "ENUM of Client Types",
        "type": "enum",
        "symbols": [
          "producer",
          "consumer"
        ]
      },
      "doc": "identifies the type of client: a 'consumer' or a 'producer'"
    },
    {
      "name": "message_count",
      "type": "int",
      "doc": "number of messages published/consumed by the client"
    },
    {
      "name": "start_timestamp",
      "type": "int",
      "doc": "Time the monitoring system started counting messages"
    },
    {
      "name": "host_info",
      "type": "string",
      "doc": "Host information of client"
    }
  ]
}
