# -*- coding: utf-8 -*-
# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import
from __future__ import unicode_literals

import glob


def get_file_paths_from_glob_patterns(glob_patterns):
    """ Return a set of files matching the given list of glob patterns
     (for example ["./test.sql", "./other_tables/*.sql"])
    """
    file_paths = set()
    for glob_pattern in glob_patterns:
        file_paths |= set(glob.glob(glob_pattern))
    return file_paths
